/*   fsmbuild
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/*	This file contains generic material which needs to be included
	in any finite state machine generated by fsmbuild.  Some functions
	aren't required for all parsers, but they're short enough
	to include in all parsers, and I strongly prefer to make them
	inline.  */

/*	Include appropriate generic headers.  */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>

#if STDC_HEADERS
#include <stdlib.h>
#endif

#if WITH_DMALLOC
#include "dmalloc.h"
#endif


/*	The fsm_stack contains a stack of states; this is used so that
	you can return to a state.  */
struct fsm_stack {
	int state;
	struct fsm_stack *next;
};

/*	The struct fsm_state holds all state infomation needed by a parser.  */
struct fsm_state {
	/*	The number of the state to use next time through the loop.  */
	int state;

	/*	A stack of states.  */
	struct fsm_stack *stack;

	/*	A string which is accumulated during parsing. */
	char *string;

	/*	The size of `string' as passed to malloc.  */
	size_t string_size;

	/*	Number of actual valid bytes in `string'.  */
	size_t string_count;
};

/*	Preform initialization for a struct fsm_state.  */
inline static void fsm_clear_state(struct fsm_state *m)
{
	m->state = 0;
	m->stack = 0;
	m->string_size = 256;
	m->string_count = 0;
	m->string = malloc(m->string_size);
}

/*	allocate a new fsm_state with `malloc'.  */
inline static struct fsm_state *fsm_new_state(void)
{
	struct fsm_state *m;

	m = malloc(sizeof(struct fsm_state));
	fsm_clear_state(m);
	return m;
}

/*	free all data structures referenced by `m', but don't free `m'
	itself.  */
inline static void fsm_cleanup_state(struct fsm_state *m)
{
	struct fsm_stack *stack;

	free(m->string);
	while (m->stack) {
		stack = m->stack;
		m->stack = m->stack->next;
		free(stack);
	}
}

/*	free `m' and all the data structures `m' refers to.  */
inline static void fsm_free_state(struct fsm_state *m)
{
	fsm_cleanup_state(m);
	free(m);
}

/*	push a state onto the stack  */
inline static void fsm_push_state(struct fsm_state *m, int state)
{
	struct fsm_stack *n;

	n = malloc(sizeof(struct fsm_stack));
	n->next = m->stack;
	n->state = state;
	m->stack = n;
}

/*	pop a state off the stack  */
inline static int fsm_pop_state(struct fsm_state *m)
{
	struct fsm_stack *n;
	int y_ask_y;

	n = m->stack;
	m->stack = n->next;
	y_ask_y = n->state;
	free(n);
	return y_ask_y;
}

/*	clear the string stored in the `m' */
inline static void fsm_clear_string(struct fsm_state *m)
{
	m->string_count = 0;
}

/*	append `c' to `m' */
inline static void fsm_append_char(struct fsm_state *m, char c)
{
	m->string[m->string_count] = c;
	m->string_count++;

	if (m->string_count == m->string_size) {
		m->string_size *= 2;
		m->string = realloc(m->string, m->string_size);
	}
}

/*	append the null-terminated string `s' to the string stored in `m'.  */
inline static void fsm_append_null_string(struct fsm_state *m, char **s)
{
	size_t count;

	if (*s) {
		count = strlen(*s);
		*s = realloc(*s, m->string_count + count + 1);
		memcpy((*s) + count, m->string, m->string_count);
		(*s)[m->string_count + count] = 0;
	} else {
		*s = malloc(m->string_count + 1);
		memcpy(*s, m->string, m->string_count);
		(*s)[m->string_count] = 0;
	}
}

/*	append the string `s' of size `count' to the string stored in `m'  */
inline static void fsm_append_sized_string(struct fsm_state *m, char **s,
	size_t *count)
{
	if ((*s) && (*count)) {
		*s = realloc(*s, *count + m->string_count);
		memcpy((*s) + (*count), m->string, m->string_count);
		*count += m->string_count;
	} else {
		*s = malloc(m->string_count);
		memcpy(*s, m->string, m->string_count);
		*count = m->string_count;
	}
}

/*	Do a case-insensitive comparision between the string in
	`m' and the null-terminated string pointed to by `s'.
	returns the same values as `strncasecmp'  */
/*	FIXME: this loses if you don't have strncasecmp, or if you
	have null bytes.  */
inline static int fsm_case_cmp(struct fsm_state *m, char *s)
{
	size_t count;

	count = strlen(s);

	if (count != m->string_count)
		return -1;

	return strncasecmp(m->string, s, count);
}

/*	a kludge to make m->string a valid, null-terminated string.  */
inline static char *fsm_tmp_null_string(struct fsm_state *m)
{
	m->string[m->string_count] = 0;
	return m->string;
}

