/*   fsmbuild
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include <sys/types.h>
#include <getopt.h>
#include <stdio.h>
#include <ctype.h>

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

/* Allow strings to be marked for gettext without bothering to
   use it yet.  */
#define _(foo) (foo)

/* The name of the program.  */
char *progname;

/* Generate #line directives if true.  */
int do_line = 1;

/* Name of current input file. */
char *filename;

/* Name of the output file. */
char *outfile;

/* The line number within the input file. */
int line_number = 0;

/* The exit status, stored for the eventual exit, so that we can
   recover from errors but still warn programs like `make' of
   the error.  */
int exit_status = 0;

/* FIXME comment this */
int in_case = 0;
int in_code = 0;

enum section_enum
{
  IGNORE,
  TOP,
  MIDDLE,
  BOTTOM,
  FSM,
  PRAGMA
} section;

char *case_name;  

/* For files which have multibyte integers, the endianness of the
   input file.  */
int big_endian = 1;

/* FIXME */
int state_num = 0;

/* Storate for sections of the output file.  */
struct section {
  /* pointer to the first char of the section */
  char *data;

  /* number of valid bytes */
  size_t data_count;

  /* size passed to malloc */
  size_t data_size;
};

/* FIXME and initialize these */
struct section top;
struct section state_defines;
struct section middle;
struct section fsm_func;
struct section bottom;



void
suggest_asking_for_help (void)
{
  fprintf(stderr, _("\tTry `%s --help' for a complete list of options.\n"),
	  progname);
  exit (1);
}

void print_help (void)
{
  printf (_("%s [OPTION]... [INPUT-FILE]\n\
  Generate a finite state machine parser.
\n\
Options:\n\
-h, --help            Display this help and exit.\n\
-o, --output=FILE     Specify output filename.\n\
-v, --version         Display version information and exit.\n\
-l, --no-line         Don't generate #line directives.\n\
\n\
Email bug reports to bug-fsmbuild@gnu.ai.mit.edu.\n\
"), progname);
}

struct option longopts[] =
{
  {"help",	no_argument,		NULL,	'h'},
  {"no-line",	no_argument,		NULL,	'n'},
  {"output",	required_argument,	NULL,	'o'},
  {"version",	no_argument,		NULL,	'v'},
  {0}
};



void
print_error (char *err)
{
  exit_status = 1;
  fprintf(stderr, "%s:%d: %s\n", filename, line_number, err);
}


void
section_append (struct section *sec, char *s)
{
  size_t len;

  if (!(sec->data_size))
    {
      sec->data_size = 4096;
      sec->data = malloc (4096);
    }

  len = strlen(s);
  if (sec->data_size < (sec->data_count + len)) 
    {
      sec->data_size *= 2;
      if (sec->data_size < (sec->data_count + len)) 
	{
	  sec->data_size += len;
	}
      sec->data = realloc (sec->data, sec->data_size);
    }
  memcpy ((sec->data + sec->data_count), s, len);
  sec->data_count += len;
}

void
section_append_number (struct section *sec, int n)
{
  char s[256];
  sprintf (s, "%d", n);
  section_append (sec, s);
}


void
internal_lined (struct section *sec, int n)
{
  section_append (sec, "#line ");
  section_append_number (sec, n);
  section_append (sec, " \"");
  section_append (sec, filename);
  section_append (sec, "\"\n");
}

void
lined (struct section *sec)
{
  internal_lined (sec, line_number);
}

void
lined_early (struct section *sec)
{
  internal_lined (sec, line_number + 1);
}

void
start_state (char *state_name)
{
  if (*state_name)
    {
      section_append (&state_defines, "#define ");
      section_append (&state_defines, state_name);
      section_append (&state_defines, " ");
      section_append_number (&state_defines, state_num);
      section_append (&state_defines, "\n");
    }
  lined (&fsm_func);
  section_append (&fsm_func, "case ");
  section_append_number (&fsm_func, state_num);
  section_append (&fsm_func, ":\n");
  state_num++;
}

void
goto_state (char *state)
{
  lined (&fsm_func);
  if (!strcmp (state, "return"))
    {
      section_append (&fsm_func, "m->state = fsm_pop_state(m);\n");
    }
  else
    {
      section_append (&fsm_func, "m->state = ");
      if (*state)
	{
	  section_append (&fsm_func, state);
	}
      else
	{
	  section_append_number (&fsm_func, state_num);
	}
      section_append (&fsm_func, ";\n");
    }
}

void
finish_state (void)
{
  section_append (&fsm_func, "break;\n");
}

void
start_case (void)
{
  section_append (&fsm_func, "switch (*data) {\n");
}

void
finish_case (void)
{
  section_append (&fsm_func, "}\n");
}

void
generate_one_case_label (char *str)
{
  lined (&fsm_func);
  section_append (&fsm_func, "case ");
  section_append (&fsm_func, str);
  section_append (&fsm_func, ":\n");
}

void
generate_one_case_label_char (char c)
{
  char str [4];

  str [0] = '\'';
  str [1] = c;
  str [2] = '\'';
  str [3] = 0;
  generate_one_case_label (str);
}

void
expand_pseudo_regex (char *str)
{
  char c;
  size_t len = strlen(str);
  if ((*str == '[') && (str [len - 1] == ']'))
    {
      str++;
      while (*(str + 1))
	{
	  if ((*str) && (str [1] == '-'))
	    {
	      for (c = str [0]; c <= str [2]; c++)
		{
		  generate_one_case_label_char (c);
		}
	      str += 3;
	    }
	  else
	    {
	      generate_one_case_label_char (*str);
	      str++;
	    }
	}
    }
  else
    {
      generate_one_case_label (str);
    }
}

void
split_case_labels_on_commas (char *str)
{
  char *s, *orig_s;
  char *token;

  s = orig_s = strdup (str);

  token = s;
  while (*s)
    {
      if (*s == '\'')
	{
	  s++;
	  while (*s && (*s != '\''))
	    {
	      if (*s == '\\')
		s++;
	      s++;
	    }
	}
      if (*s == ',')
	{
	  *s = 0;
	  expand_pseudo_regex (token);
	  token = s + 1;
	}
      s++;
    }
  expand_pseudo_regex (token);
  free(orig_s);
}

void
case_line (char *one, char *two, char *three, char *four, char *five)
{
  if (*two)
    {
      print_error ("Can't use second field");
    }

  if (!strcmp (one, "default"))
    {
      lined (&fsm_func);
      section_append (&fsm_func, "default:\n");
    }
  else
    {
      split_case_labels_on_commas (one);
    }

  if (*three)
    {
      lined (&fsm_func);
      if (!strcmp (three, "%"))
	{
	  section_append (&fsm_func, "fsm_append_char (m, (*data));\n");
	}
      else
	{
	  section_append (&fsm_func, "x->");
	  section_append (&fsm_func, three);
	  section_append (&fsm_func, " = (*data);\n");
	}
    }

  goto_state(four);

  if (!strcmp (five, "%"))
    {
      in_code = 1;
    }
  else if (!strcmp (five, "continue"))
    {
      lined (&fsm_func);
      section_append (&fsm_func, "continue;\n");
      finish_state ();
    }
  else
    {
      finish_state ();
      if (*five)
	print_error ("Don't know what to do with fifth field");
    }
}      


int
is_numeric_string (char *s)
{
  while (*s)
    {
      if (!isdigit (*s))
	return 0;
      s++;
    }
  return 1;
}

int
is_alphanumeric_string (char *s)
{
  if (!isalpha (*s))
    return 0;
  s++;
  while (*s)
    {
      if (!(isdigit (*s) || isalpha (*s) || (*s == '_')))
	return 0;
      s++;
    }
  return 1;
}

int
string_to_int (char *s)
{
  int i = 0;

  while (*s) {
    i *= 10;
    i += *s - '0';
    s++;
  }
  return i;
}

void
handle_five (char *five)
{
  if (!strcmp (five, "%"))
    {
      in_code = 1;
    }
  else if (!strcmp (five, "continue"))
    {
      section_append (&fsm_func, "continue;\n");
      finish_state ();
    }
  else
    {
      finish_state ();
      if (*five)
	{
	  print_error ("Don't know what to do with fifth field");
	}
    }
}

void
numeric_state_line (char *one, int count, char *three, char *four, char *five)
{
  int i, n;

  if ((*three) && (strcmp (three, "%"))) 
    {
      lined (&fsm_func);
      section_append (&fsm_func, "x->");
      section_append (&fsm_func, three);
      section_append (&fsm_func, " = 0;\n");
    }

  for (i = 0; i < count; i++)
    {
      if (i)
	{
	  start_state ("");
	}
      if (!strcmp (three, "%"))
	{
	  lined (&fsm_func);
	  section_append (&fsm_func, "fsm_append_char(m, (*data));\n");
	}
      else if (*three)
	{
	  if (big_endian)
	    {
	      n = (count - i - 1) * 8;
	    }
	  else
	    {
	      n = (i) * 8;
	    }
	  lined (&fsm_func);
	  section_append (&fsm_func, "x->");
	  section_append (&fsm_func, three);
	  section_append (&fsm_func, " += (*data) << ");
	  section_append_number (&fsm_func, n);
	  section_append (&fsm_func, ";\n");
	}
      if ((i + 1) == count)
	{
	  goto_state (four);
	}
      else
	{
	  goto_state ("");
	}
      handle_five (five);
    }
}


void
alphanumeric_state_line (char *one, char *two, char *three, char *four,
			 char *five)
{
  lined (&fsm_func);
  section_append (&fsm_func, "if (!(");
  if (is_alphanumeric_string (two)) {
    section_append (&fsm_func, "x->");
  }
  section_append (&fsm_func, two);
  section_append (&fsm_func, ")) {\n");
  goto_state (four);
  section_append (&fsm_func, "}\n");
  handle_five (five);
  if (*three)
    {
      print_error ("Don't know what to do with third field");
    }
}

void
state_line (char *one, char *two, char *three, char *four, char *five)
{
  start_state (one);

  if (!*two) {
    numeric_state_line (one, 1, three, four, five);
    return;
  }

  if (is_numeric_string (two))
    {
      numeric_state_line (one, string_to_int (two), three, four, five);
    }
  else
    {
      alphanumeric_state_line (one, two, three, four, five);
    }
}

void
handle_tokenized_fsm_line (char *one, char *two, char *three, char *four,
			   char *five)
{
  if (in_case)
    {
      case_line (one, two, three, four, five);
    }
  else
    {
      state_line (one, two, three, four, five);
    }
}

/* Strip linear whitespace by returning a pointer to the first non-lws
   character and inserting a zero at the new end of the string.  */
char *
strip_lws (char *s, char *end)
{
  while ((*s == ' ') || (*s == '\t'))
    s++;
  *end = 0;
  end--;
  while ((*end == ' ') || (*end == '\t'))
    end--;
  end++;
  *end = 0;
  return s;
}


void
case_or_state_line (char *str)
{
  char *s, *orig_s;
  char *token;
  char *args[5];
  int count = 0;
  int done = 0;

  s = orig_s = strdup (str);

  args[0] = "";
  args[1] = "";
  args[2] = "";
  args[3] = "";
  args[4] = "";

  token = s;
  while (1)
    {
      if (*s == '\'')
	{
	  s++;
	  while (*s && (*s != '\''))
	    {
	      if (*s == '\\')
		s++;
	      s++;
	    }
	}
      if ((*s == ':') || (!*s))
	{
	  if (!*s)
	    done = 1;
	  args[count] = strip_lws (token, s);
	  count++;
	  if (count == 6)
	    {
	      print_error ("Too many arguements");
	      handle_tokenized_fsm_line (args[0], args[1], args[2],
					 args[3], args[4]);
	      free (orig_s);
	      return;
	    }
	  if (!done)
	    {
	      token = s + 1;
	    }
	  else
	    {
	      handle_tokenized_fsm_line (args[0], args[1], args[2],
					 args[3], args[4]);
	      free (orig_s);
	      return;
	    }
	}
      s++;
    }
}

void
handle_fsm_line (char *line)
{
  if (!*line)
    return;

  if (!strcmp (line, "%"))
    {
      if (in_code)
	{
	  finish_state ();
	  in_code = 0;
	}
      else if (in_case)
	{
	  finish_case ();
	  finish_state ();
	  in_case = 0;
	}
      return;
    }

  if (*line == '%')
    {
      if (in_code)
	{
	  finish_state ();
	}
      if (in_case)
	{
	  finish_case ();
	  finish_state ();
	}
      if (case_name)
	free (case_name);
      case_name = strdup (line + 1);
      in_case = 1;
      start_state (case_name);
      start_case ();
      return;
    }

  if (in_code)
    {
      section_append (&fsm_func, line);
      section_append (&fsm_func, "\n");
    }
  else
    {
      case_or_state_line (line);
    }
}


void
handle_data_line (char *line)
{
  switch (section)
    {
    case TOP:
      section_append (&top, line);
      section_append (&top, "\n");
      break;
    case MIDDLE:
      section_append (&middle, line);
      section_append (&middle, "\n");
      break;
    case BOTTOM:
      section_append (&bottom, line);
      section_append (&bottom, "\n");
      break;
    case FSM:
      handle_fsm_line (line);
      break;
    case PRAGMA:
      if (!strcmp (line, "big-endian"))
	{
	  big_endian = 1;
	}
      else if (!strcmp (line, "little-endian"))
	{
	  big_endian = 0;
	}
      else if (*line)
	{
	  print_error ("Unrecognized pragma");
	}
      break;
    }
}

void
handle_line (char *line)
{
  if (!strcmp (line, "@ignore"))
    {
      section = IGNORE;
    }
  else if (!strcmp (line, "@top"))
    {
      lined_early (&top);
      section = TOP;
    }
  else if (!strcmp (line, "@middle"))
    {
      lined_early (&middle);
      section = MIDDLE;
    }
  else if (!strcmp (line, "@bottom"))
    {
      lined_early (&bottom);
      section = BOTTOM;
    }
  else if (!strcmp (line, "@pragma"))
    {
      section = PRAGMA;
    }
  else if (!strcmp (line, "@fsm"))
    {
      section = FSM;
    }
  else if ((*line == '@') && (*(line + 1)))
    {
      print_error ("Unrecognized section");
      section = IGNORE;
    }
  else
    {
      handle_data_line (line);
    }
}

#define MY_BUF_SIZE 4096
void
read_file (void)
{
  char buf[MY_BUF_SIZE];
  char *data;
  size_t data_size, data_count, count, i;
  int fd;

  data_count = 0;
  data_size = 4096;
  data = malloc (data_size);

  if (filename) 
    {
      fd = open (filename, O_RDONLY);
      if (fd < 0)
	{
	  perror (filename);
	  exit (1);
	}
    }
  else
    fd = 0;
  while (1)
    {
      count = read (fd, buf, MY_BUF_SIZE);
      if (count <= 0)
	return;
      for (i = 0; i < count; i++)
	{
	  if (buf [i] == '\n')
	    {
	      line_number++;
	      data [data_count] = 0;
	      handle_line (data);
	      data_count = 0;
	    }
	  else
	    {
	      data [data_count] = buf [i];
	      data_count++;
	    }
	  if ((data_count + 2) == data_size) {
	    data_size *= 2;
	    data = realloc (data, data_size);
	  }
	}
    }
}


void
write_file (void)
{
  FILE *myout;

  if (outfile)
    {
      myout = fopen (outfile, "w");
    }
  else
    {
      myout = stdout;
    }

  fprintf (myout,
	   "/* Generated automagically by fsmbuild " VERSION
	   "; manual editing is strongly\n   frowned upon.  */\n\n");
  fwrite (top.data, 1, top.data_count, myout);
  fprintf (myout, "\n#include <fsm.h>\n");
  fwrite (state_defines.data, 1, state_defines.data_count, myout);
  fwrite (middle.data, 1, middle.data_count, myout);
  fprintf (myout, "while (data_size > 0) {\n"
	   "switch (m->state) {\n");
  fwrite (fsm_func.data, 1, fsm_func.data_count, myout);
  fprintf (myout, "}\n" "data++, data_size--;\n" "}\n" "}\n");
  fwrite (bottom.data, 1, bottom.data_count, myout);
}



int
main (int argc, char *argv[])
{
  progname = argv[0];

  while (1) 
    {
      int opt = getopt_long (argc, argv, "hnvo:", longopts, 0);

      if (opt == EOF)
	break;

      switch (opt) 
	{
	case 0:
	  break;

	case 'o':
	  if (outfile) 
	    {
	      fprintf (stderr, _("%s: Specify the output file only once.\n"),
		       progname);
	      suggest_asking_for_help();
	    }
	  else
	    {
	      outfile = optarg;
	    }
	  break;

	case 'l':
	  do_line = 0;
	  break;

	case 'h':
	  print_help();
	  exit(0);

	case 'v':
	  printf (_("install-info (GNU %s) %s\n"), PACKAGE, VERSION);
	  puts (_("Copyright (C) 1997 Free Software Foundation, Inc.\n\
There is NO warranty.  You may redistribute this software\n\
under the terms of the GNU General Public License.\n\
For more information about these matters, see the files named COPYING."));
	  exit (0);
	default:
	  suggest_asking_for_help();
	}
    }
  for (; optind < argc; ++optind)
    {
      if (!filename)
	filename = argv [optind];
      else
	{
	  fprintf (stderr, "extra command line arguement ignored\n");
	}
    }
  read_file ();
  write_file ();
  return exit_status;
}
